const {
  isParent, isChild, isOnlyChild,
  getBounds, getBoundsTuple, boundsOverlap, expandBounds, getBoundsCopy,
  setParent, removeParent
} = require('./util');

const addListener = function(event, selector, callback) {
  this.listeners.push({ event, selector, callback });

  if (selector == null) {
    this.cy.on(event, callback);
  } else {
    this.cy.on(event, selector, callback);
  }
};

const addListeners = function() {
  const { options, cy } = this;

  const isMultiplySelected = n => n.selected() && cy.elements('node:selected').length > 1;
  const canBeGrabbed = n => !isMultiplySelected(n) && options.grabbedNode(n);
  const canBeDropTarget = n => !n.same(this.grabbedNode) && options.dropTarget(n, this.grabbedNode);
  const canBeDropSibling = n => !n.same(this.grabbedNode) && options.dropSibling(n, this.grabbedNode);
  const canPullFromParent = n => isChild(n);
  const getBoundTuplesNode = n => getBoundsTuple(n, options.boundingBoxOptions);

  const canBeInBoundsTuple = n => (!isParent(this.grabbedNode) || !this.grabbedNode.descendants().anySame(n)) && (canBeDropTarget(n) || canBeDropSibling(n)) && !n.same(this.dropTarget);
  const updateBoundsTuples = () => this.boundsTuples = cy.nodes(canBeInBoundsTuple).map(n => {
    console.log(n.data("id"))
    return n
  }).map(getBoundTuplesNode);

  const reset = () => {
    this.grabbedNode.removeClass('cdnd-grabbed-node');
    this.dropTarget.removeClass('cdnd-drop-target');
    this.dropSibling.removeClass('cdnd-drop-sibling');

    this.grabbedNode = cy.collection();
    this.dropTarget = cy.collection();
    this.dropSibling = cy.collection();
    this.dropTargetBounds = null;
    this.boundsTuples = [];
    this.inGesture = false;
  };

  this.addListener('grab', 'node', e => {
    const node = e.target;

    if (!this.enabled || !canBeGrabbed(node)) { return; }

    this.inGesture = true;
    this.grabbedNode = node;
    this.dropTarget = cy.collection();
    this.dropSibling = cy.collection();

    console.log({ grabbedNode: this.grabbedNode.data("id") })

    // if (canPullFromParent(node) && !isChild(this.grabbedNode)) {
    //   this.dropTarget = node.parent();
    //   this.dropTargetBounds = getBoundsCopy(this.dropTarget, options.boundingBoxOptions);
    // }

    updateBoundsTuples();

    this.grabbedNode.addClass('cdnd-grabbed-node');
    this.dropTarget.addClass('cdnd-drop-target');

    node.emit('cdndgrab');
  });

  this.addListener('add', 'node', e => {
    if (!this.inGesture || !this.enabled) { return; }

    const newNode = e.target;

    if (canBeInBoundsTuple(newNode)) {
      this.boundsTuples.push(getBoundsTuple(newNode, options.boundingBoxOptions));
    }
  });

  this.addListener('remove', 'node', e => {
    if (!this.inGesture || !this.enabled) { return; }

    const rmedNode = e.target;
    const rmedIsTarget = rmedNode.same(this.dropTarget);
    const rmedIsSibling = rmedNode.same(this.dropSibling);
    const rmedIsGrabbed = rmedNode.same(this.grabbedNode);

    // try to clean things up if one of the drop nodes is removed
    if (rmedIsTarget || rmedIsSibling || rmedIsGrabbed) {
      if (rmedIsGrabbed) {
        reset();
      } else {
        this.dropTarget = cy.collection();
        this.dropSibling = cy.collection();

        updateBoundsTuples();
      }
    }
  });

  this.addListener('drag', 'node', () => {
    if (!this.inGesture || !this.enabled) { return; }

    if (this.dropTarget.nonempty()) { // already in a parent
      const bb = expandBounds(getBounds(this.grabbedNode, options.boundingBoxOptions), options.outThreshold);
      const parent = this.dropTarget;
      const sibling = this.dropSibling;
      const rmFromParent = !boundsOverlap(this.dropTargetBounds, bb);
      const grabbedIsOnlyChild = isOnlyChild(this.grabbedNode);

      if (rmFromParent) {
        removeParent(this.grabbedNode);
        removeParent(this.dropSibling);

        this.dropTarget.removeClass('cdnd-drop-target');
        this.dropSibling.removeClass('cdnd-drop-sibling');

        let isFakeParent = this.dropTarget && this.dropTarget.hasClass('cdnd-new-parent');
        if (
          (this.dropSibling.nonempty() // remove extension-created parents on out
            || grabbedIsOnlyChild) && isFakeParent// remove empty parents
        ) {
          this.dropTarget.remove();
        }

        this.dropTarget = cy.collection();
        this.dropSibling = cy.collection();
        this.dropTargetBounds = null;

        if (isFakeParent) {
          updateBoundsTuples();
        }

        this.grabbedNode.emit('cdndout', [parent, sibling]);
      }


    } else { // not in a parent
      const bb = expandBounds(getBounds(this.grabbedNode, options.boundingBoxOptions), options.overThreshold);
      const tupleOverlaps = t => !t.node.removed() && boundsOverlap(bb, t.bb);
      const overlappingNodes = this.boundsTuples.filter(tupleOverlaps).map(t => t.node);

      if (overlappingNodes.length > 0) { // potential parent
        console.log({ ovNodes: overlappingNodes.map(n => n.data("id")) })
        const overlappingParents = overlappingNodes.filter(isParent);
        let parent, sibling;

        if (overlappingParents.length > 0) {
          sibling = cy.collection();
          parent = overlappingParents[0]; // TODO maybe use a metric here to select which one
        } else {
          sibling = overlappingNodes[0]; // TODO maybe use a metric here to select which one
          parent = cy.add(options.newParentNode(this.grabbedNode, sibling));
        }

        parent.addClass('cdnd-drop-target');

        if (parent !== sibling) {
          parent.addClass('cdnd-new-parent');
        }

        sibling.addClass('cdnd-drop-sibling');

        setParent(sibling, parent);

        this.dropTargetBounds = getBoundsCopy(parent, options.boundingBoxOptions);

        setParent(this.grabbedNode, parent);

        this.dropTarget = parent;
        this.dropSibling = sibling;

        this.grabbedNode.emit('cdndover', [parent, sibling]);
      }
    }
  });

  this.addListener('free', 'node', () => {
    if (!this.inGesture || !this.enabled) { return; }

    const { grabbedNode, dropTarget, dropSibling } = this;

    reset();

    grabbedNode.emit('cdnddrop', [dropTarget, dropSibling]);
  });
};

const removeListeners = function() {
  const { cy } = this;

  this.listeners.forEach(lis => {
    const { event, selector, callback } = lis;

    if (selector == null) {
      cy.removeListener(event, callback);
    } else {
      cy.removeListener(event, selector, callback);
    }
  });

  this.listeners = [];
};

module.exports = { addListener, addListeners, removeListeners };
